package logs

import (
	"context"
	"time"

	itemstransportgrpc "git.perx.ru/perxis/perxis-go/pkg/items/transport/grpc"
	"git.perx.ru/perxis/perxis-go/pkg/options"
	pb "git.perx.ru/perxis/perxis-go/proto/logs"
	"google.golang.org/protobuf/types/known/timestamppb"
)

const ServiceName = "logs"

type Service interface {

	// Log метод записи логов
	Log(ctx context.Context, entries []*Entry) error

	// Find метод для поиска логов по заданным параметрам
	Find(ctx context.Context, filter *Filter, options *FindOptions) (*FindResult, error)

	// Delete метод для удаления логов по заданным параметрам
	Delete(ctx context.Context, filter *Filter) error
}

type Filter struct {
	Q []string
}

type FindRequest struct {
	Filter  *Filter
	Options *FindOptions
}

type FindResult struct {
	// Найденные записи лога
	Entries []*Entry
	// Использовавшийся для поиска фильтр
	Filter *Filter
	// Использовавшиеся для поиска опции
	Options *options.FindOptions
	// Запрос для получения следующей страницы
	Next *FindRequest
	// Запрос для получения предыдущей страницы
	Prev *FindRequest
}

type FindOptions struct {
	// Сортировка результатов
	Sort []string
	// Включить поля в результаты
	Fields []string
	// Исключить поля из результатов
	ExcludeFields bool
	// Ограничение количества результатов
	Limit int
	// Ограничение результатов по времени начиная с этим временем (pagination)
	From time.Time
	// Ограничение результатов по времени заканчивая этим временем (pagination)
	To time.Time
}

// FindResultToPB перевод в proto представление FindResult
func FindResultToPB(result *FindResult) *pb.FindResult {
	findResult := &pb.FindResult{Total: 0} // поле стало deprecated всегда возвращаем 0

	entries := make([]*pb.LogEntry, 0, len(result.Entries))
	for _, e := range result.Entries {
		entries = append(entries, EntryToPB(e))
	}
	findResult.Entries = entries

	if result.Filter != nil {
		findResult.Filter.Q = result.Filter.Q
	}
	if result.Options != nil {
		findResult.Options, _ = itemstransportgrpc.PtrServicesFindOptionsToProto(result.Options)
	}

	if result.Next != nil {
		findResult.Next = FindRequestToPB(result.Next)
	}

	if result.Prev != nil {
		findResult.Prev = FindRequestToPB(result.Prev)
	}

	return findResult
}

// FindResultFromPB перевод из proto представления FindResult
func FindResultFromPB(result *pb.FindResult) *FindResult {
	findResult := new(FindResult)

	entries := make([]*Entry, 0, len(result.Entries))
	for _, e := range result.Entries {
		entries = append(entries, EntryFromPB(e))
	}
	findResult.Entries = entries

	if result.Filter != nil {
		findResult.Filter.Q = result.Filter.Q
	}
	if result.Options != nil {
		findResult.Options, _ = itemstransportgrpc.ProtoToPtrServicesFindOptions(result.Options)
	}

	if result.Next != nil {
		findResult.Next = FindRequestFromPB(result.Next)
	}

	if result.Prev != nil {
		findResult.Prev = FindRequestFromPB(result.Prev)
	}

	return findResult
}

// FindRequestToPB перевод в proto представление FindRequest
func FindRequestToPB(request *FindRequest) *pb.FindRequest {
	if request == nil {
		return nil
	}

	req := new(pb.FindRequest)

	if request.Filter != nil {
		req.Filter = &pb.Filter{Q: request.Filter.Q}
	}

	if request.Options != nil {
		req.Options = FindOptionsToPB(request.Options)
	}
	return req
}

// FindRequestFromPB перевод из proto представления FindRequest
func FindRequestFromPB(request *pb.FindRequest) *FindRequest {
	if request == nil {
		return nil
	}

	req := new(FindRequest)

	if request.Filter != nil {
		req.Filter = &Filter{Q: request.Filter.Q}
	}

	if request.Options != nil {
		req.Options = FindOptionsFromPB(request.Options)
	}

	return req
}

// FindOptionsToPB перевод в proto представление FindOptions
func FindOptionsToPB(opts *FindOptions) *pb.FindOptions {
	if opts == nil {
		return nil
	}

	res := &pb.FindOptions{
		Sort:          opts.Sort,
		Fields:        opts.Fields,
		ExcludeFields: opts.ExcludeFields,
		Limit:         int32(opts.Limit),
	}

	if !opts.From.IsZero() {
		res.From = timestamppb.New(opts.From)
	}

	if !opts.To.IsZero() {
		res.To = timestamppb.New(opts.To)
	}

	return res
}

// FindOptionsFromPB перевод из proto представления FindOptions
func FindOptionsFromPB(opts *pb.FindOptions) *FindOptions {
	if opts == nil {
		return nil
	}
	res := &FindOptions{
		Sort:          opts.Sort,
		Fields:        opts.Fields,
		ExcludeFields: opts.ExcludeFields,
		Limit:         int(opts.Limit),
	}

	if opts.From.IsValid() {
		res.From = opts.From.AsTime()
	}
	if opts.To.IsValid() {
		res.To = opts.To.AsTime()
	}
	return res
}
